@@ -1,35 +1,3 @@
---
--- Tests for ‘path-io’ package.
---
--- Copyright © 2016–2017 Mark Karpov <markkarpov@openmailbox.org>
---
--- Redistribution and use in source and binary forms, with or without
--- modification, are permitted provided that the following conditions are
--- met:
---
--- * Redistributions of source code must retain the above copyright notice,
---   this list of conditions and the following disclaimer.
---
--- * Redistributions in binary form must reproduce the above copyright
---   notice, this list of conditions and the following disclaimer in the
---   documentation and/or other materials provided with the distribution.
---
--- * Neither the name Mark Karpov nor the names of contributors may be used
---   to endorse or promote products derived from this software without
---   specific prior written permission.
---
--- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS “AS IS” AND ANY
--- EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
--- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
--- DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
--- DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
--- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
--- OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
--- HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
--- STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
--- ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
--- POSSIBILITY OF SUCH DAMAGE.
-
 {-# LANGUAGE CPP             #-}
 {-# LANGUAGE TemplateHaskell #-}
 
@@ -37,6 +5,7 @@ module Main (main) where
 
 import Control.Monad
 import Control.Monad.Catch
+import Control.Monad.IO.Class (MonadIO (..))
 import Data.List (sort)
 import Path
 import Path.IO
@@ -50,7 +19,10 @@ import Control.Applicative ((<$>))
 
 main :: IO ()
 main = hspec . around withSandbox $ do
+#ifndef mingw32_HOST_OS
   beforeWith populatedDir $ do
+    -- NOTE These tests shall fail on Windows as unix-compat does not
+    -- implement createSymbolicLink for windows.
     describe "listDir"          listDirSpec
     describe "listDirRecur"     listDirRecurSpec
     describe "listDirRecurWith" listDirRecurWithSpec
@@ -58,15 +30,15 @@ main = hspec . around withSandbox $ do
     describe "copyDirRecur"     copyDirRecurSpec
     describe "copyDirRecur'"    copyDirRecur'Spec
     describe "findFile"         findFileSpec
-  -- This test may fail on windows as unix-compat does not implement
-  -- createSymbolicLink.
-#ifndef mingw32_HOST_OS
   beforeWith populatedCyclicDir $
     describe "listDirRecur Cyclic" listDirRecurCyclicSpec
 #endif
   describe "getCurrentDir"    getCurrentDirSpec
   describe "setCurrentDir"    setCurrentDirSpec
   describe "withCurrentDir"   withCurrentDirSpec
+#ifndef mingw32_HOST_OS
+  -- NOTE We can't quite test this on Windows as well, because the
+  -- environmental variables HOME and TMPDIR do not exist there.
   describe "getHomeDir"       getHomeDirSpec
   describe "getTempDir"       getTempDirSpec
 #if MIN_VERSION_directory(1,2,3)
@@ -74,6 +46,7 @@ main = hspec . around withSandbox $ do
   describe "getXdgDir Config" getXdgConfigDirSpec
   describe "getXdgDir Cache"  getXdgCacheDirSpec
 #endif
+#endif
 
 listDirSpec :: SpecWith (Path Abs Dir)
 listDirSpec = it "lists directory" $ \dir ->
@@ -92,9 +65,9 @@ listDirRecurWithSpec =
         `shouldReturn` populatedDirRecurWith
 
 listDirRecurWith
-  :: (Path Abs Dir ->  IO Bool)            -- ^ Dir match predicate
+  :: (Path Abs Dir ->  IO Bool)           -- ^ Dir match predicate
   -> (Path Abs File -> IO Bool)           -- ^ File match predicate
-  -> Path Abs Dir                        -- ^ Top dir to traverse
+  -> Path Abs Dir                         -- ^ Top dir to traverse
   -> IO ([Path Abs Dir], [Path Abs File]) -- ^ Matched subdirs and files
 listDirRecurWith dirPred filePred =
   walkDirAccum Nothing $ \_ d f -> do
@@ -102,10 +75,17 @@ listDirRecurWith dirPred filePred =
     f' <- filterM filePred f
     return (d', f')
 
+-- Follows symbolic links
+listDirRecurCyclic :: (MonadIO m, MonadThrow m)
+  => Path b Dir                          -- ^ Directory to list
+  -> m ([Path Abs Dir], [Path Abs File]) -- ^ Sub-directories and files
+listDirRecurCyclic = walkDirAccum Nothing (\_ d f -> return (d, f))
+
 listDirRecurCyclicSpec :: SpecWith (Path Abs Dir)
 listDirRecurCyclicSpec =
   it "lists directory trees having traversal cycles" $ \dir ->
-    getDirStructure listDirRecur dir `shouldReturn` populatedCyclicDirStructure
+    getDirStructure listDirRecurCyclic dir
+        `shouldReturn` populatedCyclicDirStructure
 
 -- | walkDir with a Finish handler may have unpredictable output depending on
 -- the order of traversal. The only guarantee is that we will finish only after
@@ -235,7 +215,7 @@ getXdgCacheDirSpec =
 ----------------------------------------------------------------------------
 -- Helpers
 
--- | Create sandbox directory to model some situation in it and run some
+-- | Create a sandbox directory to model some situation in it and run some
 -- tests. Note that we're using new unique sandbox directory for each test
 -- case to avoid contamination and it's unconditionally deleted after test
 -- case finishes.
@@ -250,11 +230,15 @@ withSandbox = withSystemTempDir "path-io-sandbox"
 
 populatedDir :: Path Abs Dir -> IO (Path Abs Dir)
 populatedDir root = do
-  let (dirs, files) = populatedDirStructure
+  let (_, files) = populatedDirStructure
       pdir          = root </> $(mkRelDir "pdir")
       withinSandbox = (pdir </>)
   ensureDir pdir
-  forM_ dirs (ensureDir . withinSandbox)
+  ensureDir $ withinSandbox $(mkRelDir "b")
+  ensureDir $ withinSandbox $(mkRelDir "b/c")
+  -- to verify that we do not follow symbolic links. We should not list b's
+  -- tree under 'a'.
+  createSymbolicLink "b" (toFilePath $ withinSandbox $(mkRelFile "a"))
   forM_ files $ (`writeFile` "") . toFilePath . withinSandbox
   return pdir
 
@@ -291,12 +275,12 @@ populatedDirStructure =
 -- | Create a directory structure which has cycles in it due to directory
 -- symbolic links.
 --
--- 1) Mutual cycles between two directory trees. If we traverse a or c we
+-- 1) Mutual cycles between two directory trees. If we traverse @a@ or @c@ we
 -- will get into the same cycle:
-    -- a/(b -> c), c/(d -> a)
-    -- c/(d -> a), a/(b -> c)
+--     a\/(b -> c), c\/(d -> a)
+--     c\/(d -> a), a\/(b -> c)
 -- 2) Cycle with own ancestor
-    -- e/f/(g -> e)
+--     e\/f\/(g -> e)
 
 populatedCyclicDirStructure :: ([Path Rel Dir], [Path Rel File])
 populatedCyclicDirStructure =
@@ -351,7 +335,8 @@ populatedDirRecurWith =
   ( [ $(mkRelDir "a")
     , $(mkRelDir "b")
     ]
-  , [ $(mkRelFile "b/c/three.txt")
+  , [ $(mkRelFile "a/c/three.txt") -- via symbolic link
+    , $(mkRelFile "b/c/three.txt")
     , $(mkRelFile "one.txt")
     ]
   )
